#!/usr/bin/perl -s
## $Id: bulkdns.pl,v 1.2 2002/11/11 19:06:21 stef Exp $

use strict;
use vars qw( $TIMEOUT $MAX_DNSQ $MAX_QUEUE $DNSCACHE $PIPE $VERBOSE );
use DB_File;
use Data::Dumper;
use Fcntl;


=pod

bulkdns.pl -- Program for Bulk DNS resolving in Server logfiles

Usage:
bulkdns.pl [-TIMEOUT=...] [-MAX_DNSQ=...] [-MAX_QUEUE=...] [-PIPE=...] [-VERBOSE=...] [files..]
 TIMEOUT = timeout for DNS lookups
 MAX_DNSRQ = maximal number of outstanding DNS requests at the same time (suitable
    values depend on the OS)
 MAX_QUEUE = maximal number of unresolved log entries in memory in pipe mode
 PIPE = forces pipe mode even if file is given
 VERBOSE = sum of options: 2 = statistik + 4 = dns lookups + 8 = wait in select
    if VERBOSE>1 final statistiks will be printed to STDERR
    in Summary: for unattended use use VERBOSE=1 if any, for use over dialup
    use VERBOSE=2, for local use or use over ethernet use VERBOSE=6 and if
    you are curious why it is doing nothing from time to time add +8
 files  = if no files or multiple files are given or if the only argument is '-'
    pipe mode will be used, otherwise the file will be read in two passes, first
    it resolves all IPs, next pass it replaces the IPs with the hostnames (not in place)
    This might be faster and take less memory because it is only limited by MAX_DNSRQ and not 
    by MAX_QUEUE

=cut


my %opt;
$opt{ TIMEOUT } = $TIMEOUT  || 30;
$opt{ MAX_DNSQ } = $MAX_DNSQ || 50;
$opt{ MAX_QUEUE } = $MAX_QUEUE || 5000;
my %cache; die "cannot tie cache file $DNSCACHE: $!" if ( $DNSCACHE and not
	tie %cache,'DB_File',$DNSCACHE,O_CREAT|O_RDWR,0600,$DB_BTREE);

$|=1;
if ( not $PIPE and @ARGV == 1 and $ARGV[0] ne '-' ) {
	## two passes, first get the adresses into cache, then resolve the file
	my $bdr = BulkDNSResolve->new( \%opt, undef , \%cache, $VERBOSE );
	open( F,$ARGV[0] ) || die "open $ARGV[0]: $!";
	while (<F>) { $bdr->resolve( $1 )  if ( m|^([^\s]+)(.*)| ) }
	$bdr->flush;
	$bdr = BulkDNSResolve->new( \%opt, sub { print join("",@_ ),"\n" }, \%cache );
	seek(F,0,0);
	while (<F>) { $bdr->resolve( $1,$2 )  if ( m|^([^\s]+)(.*)| ) }
	$bdr->flush;
} else {
	## use it as a pipe
	my $bdr = BulkDNSResolve->new( \%opt, sub { print join("",@_ ),"\n" }, \%cache, $VERBOSE );
	while (<>) { $bdr->resolve( $1,$2 )  if ( m|^([^\s]+)(.*)| ) }
	$bdr->flush;
}



## ------------------------------------------------------------------------------------------




package BulkDNSResolve;
use Net::DNS::Resolver;
use Net::DNS::Packet;
use strict;

use constant TIMEOUT => 0;
use constant MAX_DNSQ => 1;
use constant MAX_QUEUE => 2;
use constant RESOLVER => 3;
use constant DNSRQ => 4;
use constant QUERIES => 5;
use constant OUT => 6;
use constant CACHE => 7;
use constant RQCACHE => 8;
use constant STAT => 9;
use constant VERBOSE => 10;

use constant ST_TOTAL => 0;
use constant ST_FAILED => 1;
use constant ST_DONE => 2;
use constant ST_DNS => 3;
use constant ST_START => 4;
use constant ST_DNSOK => 5;
use constant ST_MAXQ => 6;
use constant ST_MAXDNS => 7;

use constant VERB_STAT => 2;
use constant VERB_DNS => 4;
use constant VERB_WAIT => 8;

sub DESTROY { my $self = shift; $self->flush if ( @{ $self->[ QUERIES ] } ) }
sub new {
	my ($class,$opt,$outfunc,$hash,$verbose) = @_;
	my @default_opt = ( 30,50,5000 );
	my $self = bless [ @default_opt ],$class;
	$self->[ TIMEOUT ] = delete $opt->{ 'TIMEOUT' } if ( exists $opt->{ 'TIMEOUT' });
	$self->[ MAX_DNSQ ] = delete $opt->{ 'MAX_DNSQ' } if ( exists $opt->{ 'MAX_DNSQ' });
	$self->[ MAX_QUEUE ] = delete $opt->{ 'MAX_QUEUE' } if ( exists $opt->{ 'MAX_QUEUE' });
	die "unknown options ".join(" ",keys %$opt) if (keys %$opt);
	$self->[ RESOLVER ] = Net::DNS::Resolver->new;
	$self->[ OUT ] = $outfunc;
	$self->[ CACHE ] = $hash || {};
	$self->[ RQCACHE ] = {};
	$self->[ QUERIES ] = [];
	$self->[ VERBOSE ] = $verbose;
	$self->[ STAT ]->[ ST_START ] = time();
	$self;
}

use constant STATE_INRESOLVE => 1;
use constant STATE_DUPLICATE => 2;
use constant STATE_DONE => 3;

sub resolve {
	my ($self,$host,@arg) = @_;
	$self->[ STAT ]->[ ST_TOTAL ]++;
	if ( $host =~ m|^(\d+\.\d+\.\d+\.\d+)$| ) {
		my $cache = $self->[ CACHE ]->{ $host };
		if ( $cache ) {
			if ( $self->[ DNSRQ ] ) {
				if ( $self->[ OUT ] ) {
					$self->addQ( [ $cache,[ @arg ], STATE_DONE ] ) 
				} else {
					$self->[ STAT ]->[ ST_DONE ]++;
				}
			} else {
				if ( $self->[ OUT ] ) {
					$self->shiftHead;
					&{ $self->[ OUT ] }( $cache,@arg );
				}
				$self->[ STAT ]->[ ST_DONE ]++;
			}
		} elsif ( $self->[ RQCACHE ]->{ $host } ) {
			if ( $self->[ OUT ] ) {
				$self->addQ( [ $host,[ @arg ], STATE_DUPLICATE ] ) 
			} else {
				$self->[ STAT ]->[ ST_DONE ]++;
			}
		} else {
			$self->[ DNSRQ ]++;
			$self->[ STAT ]->[ ST_MAXDNS ] = $self->[ DNSRQ ] if ( $self->[ DNSRQ ]>$self->[ STAT ]->[ ST_MAXDNS ]);
			$self->[ STAT ]->[ ST_DNS ]++;
			$self->[ RQCACHE ]->{ $host } =1;
			$self->addQ( [ $host,[ @arg ], STATE_INRESOLVE,time(),$self->[ RESOLVER ]->bgsend( $host ) ] );
		}
	} elsif ( $self->[ DNSRQ ] ) {
		if ( $self->[ OUT ] ) {
			$self->addQ( [ $host,[ @arg ], STATE_DONE ] ) 
		} else {
			$self->[ STAT ]->[ ST_DONE ]++;
		}
	} elsif ( $self->[ OUT ] ) {
		$self->shiftHead;
		$self->[ STAT ]->[ ST_DONE ]++;
		&{ $self->[ OUT ] }( $host,@arg );
	} else {
		$self->[ STAT ]->[ ST_DONE ]++;
	}
}



sub addQ {
	my ($self,$rq) = @_;
	push @{ $self->[ QUERIES ] },$rq;
	$self->[ STAT ]->[ ST_MAXQ ] = @{ $self->[ QUERIES ] } if ( @{ $self->[ QUERIES ] }>$self->[ STAT ]->[ ST_MAXQ ]);
	while ( $self->[ DNSRQ ] > $self->[ MAX_DNSQ ] or @{ $self->[ QUERIES ] } > $self->[ MAX_QUEUE ] ) {
		$self->doquery
	}
}

sub flush {
	my $self = shift;
	while ( @{ $self->[ QUERIES ] } ) { $self->doquery(1) }
	printf STDERR "stat-total time=%d queries=%d(%d) dnsrq=%d(%d) -- failed=%d ok=%d\n",
		time()-$self->[ STAT ]->[ ST_START ],
		$self->[ STAT ]->[ ST_TOTAL ],$self->[ STAT ]->[ ST_MAXQ ],
		$self->[ STAT ]->[ ST_DNS ], $self->[ STAT ]->[ ST_MAXDNS ],
		$self->[ STAT ]->[ ST_FAILED ],
		$self->[ STAT ]->[ ST_DNSOK ] if ( $self->[ VERBOSE ] );
}
	
	

sub doquery {
	my ($self,$flush) = @_;
	my @dnsrq; foreach (@{ $self->[ QUERIES ] }) {
		push @dnsrq, $_ if ( $_->[2] == STATE_INRESOLVE );
	}


	while (1) {
		my $maxwait = $self->[ TIMEOUT ];
		my $ctime = time();
		my ($failed,$succ,$done,@_dnsrq); foreach (@dnsrq) {
			my ($host,$info,$status,$time,$socket) = @$_;
			if ( $self->[ RESOLVER ]->bgisready( $_->[4] )) {
				$done++;
				$_->[2] = STATE_DONE;
				$self->[ DNSRQ ]--;
				my $packet = $self->[ RESOLVER ]->bgread( $_->[4] );
				unless ( $packet ) { 
					$failed++; 
					$self->[ STAT ]->[ ST_FAILED ]++;
					printf STDERR "dns(%d) %s failed - no packet\n",$ctime-$time,$host if ( $self->[ VERBOSE ] & VERB_DNS );
					goto CLEANUP;
				}
				my @a = $packet->answer;
				unless ( @a ) { 
					$failed++; 
					$self->[ STAT ]->[ ST_FAILED ]++;
					printf STDERR "dns(%d) %s failed - empty answer\n",$ctime-$time,$host if ( $self->[ VERBOSE ] & VERB_DNS );
					goto CLEANUP;
				}
				($_->[0]) = $a[0]->rdatastr =~m|([^\s]+)\.$|;
				$self->[ STAT ]->[ ST_DNSOK ]++; 
				$succ++;
				printf STDERR "dns(%d) %s -> %s\n", $ctime-$time, $host,$_->[0] if ( $self->[ VERBOSE ] & VERB_DNS );
			} elsif ( $ctime-$time >= $self->[ TIMEOUT ] ) {
				$done++;
				$_->[2] = STATE_DONE;
				$self->[ DNSRQ ]--;
				$self->[ STAT ]->[ ST_FAILED ]++;
				$failed++;
				printf STDERR "dns(%d) %s timeout\n",$ctime-$time,$host if ( $self->[ VERBOSE ] & VERB_DNS );
			} else {
				my $diff = $time + $self->[ TIMEOUT ] - $ctime;
				$maxwait = $diff if ( $diff < $maxwait );
				push @_dnsrq,$_;
				next; ## no CLEANUP yet
			}

			CLEANUP:
			$self->[ CACHE ]->{ $host } = ( $_->[0] ||=$host );
			delete $self->[ RQCACHE ]->{ $host };
			$socket = $_->[4] = undef;
		}

	
		$self->shiftHead if ($done);
		die "\@_dnsrq=".($#_dnsrq+1)." but DNSRQ=".$self->[ DNSRQ ] if ( @_dnsrq != $self->[ DNSRQ ] );
		printf STDERR "stat +%d query=%d/%d dnsrq=%d/%d -- failed=%d/%d ok=%d/%d done=%d\n",
			$ctime-$self->[ STAT ]->[ ST_START ],
			$self->[ STAT ]->[ ST_TOTAL ], $#{ $self->[ QUERIES ] }+1,
			$self->[ STAT ]->[ ST_DNS ], $self->[ DNSRQ ],
			$self->[ STAT ]->[ ST_FAILED ], $failed, 
			$self->[ STAT ]->[ ST_DNSOK ], $succ, 
			$done if ( $self->[ VERBOSE ] & VERB_STAT );

		last unless ( $self->[ DNSRQ ] );
		last unless  ( $flush or $self->[ DNSRQ ] > $self->[ MAX_DNSQ ] or @{ $self->[ QUERIES ] } > $self->[ MAX_QUEUE ] );
		my $vec; foreach (@_dnsrq) { vec( $vec,$_->[4]->fileno,1 ) =1 }
		print STDERR "doquery select wait=$maxwait\n" if ( $self->[ VERBOSE ] & VERB_WAIT );
		my ($rout,$eout); select( $rout=$vec,undef,$eout=$vec,$maxwait );
		@dnsrq = @_dnsrq;
	}
}

sub shiftHead {
	my $self = shift;
	while (@{ $self->[ QUERIES ] }) {
		my $top = $self->[ QUERIES ][0];
		if ( $top->[2] == STATE_DUPLICATE ) {
			my $h = $self->[ CACHE ]->{ $top->[0] };
			$top->[2] = STATE_DONE;
			$top->[0] = $h;
		}
		last unless ( $top->[2] == STATE_DONE );
		my $q = shift @{ $self->[ QUERIES ] };
		if ( $self->[ OUT ]) {
			&{$self->[ OUT ]}( $q->[0],@{ $q->[1] } );
			$self->[ STAT ]->[ ST_DONE ]++;
		}
	}
	if ( @{ $self->[ QUERIES ] } and not $self->[ DNSRQ ] ) {
		die "queries but no dnsrq, first query status=".$self->[ QUERIES ][0][2];
	}
}

	


				
			
	


__END__
 vim: set ts=4 ai:

